/*************************************************************************
* Copyright (C) 2016 Intel Corporation
*
* Licensed under the Apache License,  Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
* 	http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law  or agreed  to  in  writing,  software
* distributed under  the License  is  distributed  on  an  "AS IS"  BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the  specific  language  governing  permissions  and
* limitations under the License.
*************************************************************************/

/*
//
//  Purpose:
//     Cryptography Primitive.
//     Security Hash Standard
//     Generalized Functionality
//
//  Contents:
//        ippsHashUnpack_rmf()
//
*/

#include "owndefs.h"
#include "owncp.h"
#include "hash/pcphash_rmf.h"
#include "pcptool.h"

/*F*
//    Name: ippsHashUnpack_rmf
//
// Purpose: Unpack buffer content into the initialized context.
//
// Returns:                Reason:
//    ippStsNullPtrErr        pBuffer == NULL
//                            pState == NULL
//    ippStsContextMatchErr   pState has wrong context
//    ippStsNoErr             no errors
//
// Parameters:
//    pBuffer     pointer to the source buffer
//    pState      pointer hash state
//
*F*/
IPPFUN(IppStatus, ippsHashUnpack_rmf, (const Ipp8u* pBuffer, IppsHashState_rmf* pState))
{
    /* test pointers */
    IPP_BAD_PTR2_RET(pState, pBuffer);

    // check state
    IPP_BADARG_RET(!HASH_VALID_ID(pState, idCtxHash), ippStsContextMatchErr);

    int context_size = 0;
    ippsHashGetSizeOptimal_rmf(&context_size, HASH_METHOD(pState));
    // copy data from the buffer except the method pointer
    const IppsHashMethod* pMethod = HASH_METHOD(pState);
    CopyBlock(pBuffer, pState, context_size);
    HASH_METHOD(pState) = pMethod;

    HASH_SET_ID(pState, idCtxHash);

    /* setup pointers to buffer and hash */
    HASH_SETUP_POINTERS(pState);

    return ippStsNoErr;
}
