EXPORTS

ippcpInit
cpGetReg
cpStartTscp
cpStopTscp
cpStartTsc
cpStopTsc
cpGetCacheSize
cpGetFeature
ippcpSetCpuFeatures
ippcpGetCpuFeatures
ippcpGetCpuClocks
ippcpSetNumThreads
ippcpGetNumThreads
ippcpGetEnabledCpuFeatures
ippcpGetEnabledNumThreads
ippcpGetStatusString

ippcpGetLibVersion
cryptoGetLibVersion
ippsDESGetSize
ippsDESInit
ippsDESPack
ippsDESUnpack
ippsTDESEncryptECB
ippsTDESDecryptECB
ippsTDESEncryptCBC
ippsTDESDecryptCBC
ippsTDESEncryptCFB
ippsTDESDecryptCFB
ippsTDESEncryptOFB
ippsTDESDecryptOFB
ippsTDESEncryptCTR
ippsTDESDecryptCTR
ippsAESGetSize
ippsAESInit
ippsAESSetKey
ippsAESPack
ippsAESUnpack
ippsAESSetupNoise
ippsAES_GCMSetupNoise
ippsAES_CMACSetupNoise
ippsAESEncryptECB
ippsAESDecryptECB
ippsAESEncryptCBC
ippsAESEncryptCBC_CS1
ippsAESEncryptCBC_CS2
ippsAESEncryptCBC_CS3
ippsAESDecryptCBC
ippsAESDecryptCBC_CS1
ippsAESDecryptCBC_CS2
ippsAESDecryptCBC_CS3
ippsAESEncryptCFB
ippsAESDecryptCFB
ippsAESEncryptOFB
ippsAESDecryptOFB
ippsAESEncryptCTR
ippsAESDecryptCTR
ippsAESEncryptXTS_Direct
ippsAESDecryptXTS_Direct
ippsAES_EncryptCFB16_MB
ippsSMS4GetSize
ippsSMS4Init
ippsSMS4SetKey
ippsSMS4EncryptECB
ippsSMS4DecryptECB
ippsSMS4EncryptCBC
ippsSMS4EncryptCBC_CS1
ippsSMS4EncryptCBC_CS2
ippsSMS4EncryptCBC_CS3
ippsSMS4DecryptCBC
ippsSMS4DecryptCBC_CS1
ippsSMS4DecryptCBC_CS2
ippsSMS4DecryptCBC_CS3
ippsSMS4EncryptCFB
ippsSMS4DecryptCFB
ippsSMS4EncryptOFB
ippsSMS4DecryptOFB
ippsSMS4EncryptCTR
ippsSMS4DecryptCTR
ippsSMS4_CCMGetSize
ippsSMS4_CCMInit
ippsSMS4_CCMMessageLen
ippsSMS4_CCMTagLen
ippsSMS4_CCMStart
ippsSMS4_CCMEncrypt
ippsSMS4_CCMDecrypt
ippsSMS4_CCMGetTag
ippsAES_CCMGetSize
ippsAES_CCMInit
ippsAES_CCMMessageLen
ippsAES_CCMTagLen
ippsAES_CCMStart
ippsAES_CCMEncrypt
ippsAES_CCMDecrypt
ippsAES_CCMGetTag
ippsAES_GCMGetSize
ippsAES_GCMInit
ippsAES_GCMReinit
ippsAES_GCMReset
ippsAES_GCMProcessIV
ippsAES_GCMProcessAAD
ippsAES_GCMStart
ippsAES_GCMEncrypt
ippsAES_GCMDecrypt
ippsAES_GCMGetTag
ippsAES_XTSGetSize
ippsAES_XTSInit
ippsAES_XTSEncrypt
ippsAES_XTSDecrypt
ippsAES_S2V_CMAC
ippsAES_SIVEncrypt
ippsAES_SIVDecrypt
ippsAES_CMACGetSize
ippsAES_CMACInit
ippsAES_CMACUpdate
ippsAES_CMACFinal
ippsAES_CMACGetTag
ippsARCFourCheckKey
ippsARCFourGetSize
ippsARCFourInit
ippsARCFourReset
ippsARCFourPack
ippsARCFourUnpack
ippsARCFourEncrypt
ippsARCFourDecrypt
ippsHashGetSize
ippsHashInit
ippsHashPack
ippsHashUnpack
ippsHashDuplicate
ippsHashUpdate
ippsHashGetTag
ippsHashFinal
ippsHashMessage
ippsHashMethodGetSize
ippsHashMethodSet_MD5
ippsHashMethodSet_SM3
ippsHashMethodSet_SM3_NI
ippsHashMethodSet_SM3_TT
ippsHashMethodSet_SHA1
ippsHashMethodSet_SHA1_NI
ippsHashMethodSet_SHA1_TT
ippsHashMethodSet_SHA256
ippsHashMethodSet_SHA256_NI
ippsHashMethodSet_SHA256_TT
ippsHashMethodSet_SHA224
ippsHashMethodSet_SHA224_NI
ippsHashMethodSet_SHA224_TT
ippsHashMethodSet_SHA512
ippsHashMethodSet_SHA512_NI
ippsHashMethodSet_SHA512_TT
ippsHashMethodSet_SHA384
ippsHashMethodSet_SHA384_NI
ippsHashMethodSet_SHA384_TT
ippsHashMethodSet_SHA512_256
ippsHashMethodSet_SHA512_256_NI
ippsHashMethodSet_SHA512_256_TT
ippsHashMethodSet_SHA512_224
ippsHashMethodSet_SHA512_224_NI
ippsHashMethodSet_SHA512_224_TT
ippsHashMethodSet_SHA3_224
ippsHashMethodSet_SHA3_256
ippsHashMethodSet_SHA3_384
ippsHashMethodSet_SHA3_512
ippsHashMethodSet_SHAKE128
ippsHashMethodSet_SHAKE256
ippsHashStateMethodSet_SM3
ippsHashStateMethodSet_SM3_NI
ippsHashStateMethodSet_SM3_TT
ippsHashStateMethodSet_SHA256
ippsHashStateMethodSet_SHA256_NI
ippsHashStateMethodSet_SHA256_TT
ippsHashStateMethodSet_SHA224
ippsHashStateMethodSet_SHA224_NI
ippsHashStateMethodSet_SHA224_TT
ippsHashStateMethodSet_SHA512
ippsHashStateMethodSet_SHA512_NI
ippsHashStateMethodSet_SHA512_TT
ippsHashStateMethodSet_SHA384
ippsHashStateMethodSet_SHA384_NI
ippsHashStateMethodSet_SHA384_TT
ippsHashStateMethodSet_SHA512_256
ippsHashStateMethodSet_SHA512_256_NI
ippsHashStateMethodSet_SHA512_256_TT
ippsHashStateMethodSet_SHA512_224
ippsHashStateMethodSet_SHA512_224_NI
ippsHashStateMethodSet_SHA512_224_TT
ippsHashStateMethodSet_SHA3_224
ippsHashStateMethodSet_SHA3_256
ippsHashStateMethodSet_SHA3_384
ippsHashStateMethodSet_SHA3_512
ippsHashStateMethodSet_SHAKE128
ippsHashStateMethodSet_SHAKE256
ippsHashMethod_MD5
ippsHashMethod_SM3
ippsHashMethod_SM3_NI
ippsHashMethod_SM3_TT
ippsHashMethod_SHA1
ippsHashMethod_SHA1_NI
ippsHashMethod_SHA1_TT
ippsHashMethod_SHA256
ippsHashMethod_SHA256_NI
ippsHashMethod_SHA256_TT
ippsHashMethod_SHA224
ippsHashMethod_SHA224_NI
ippsHashMethod_SHA224_TT
ippsHashMethod_SHA512
ippsHashMethod_SHA512_NI
ippsHashMethod_SHA512_TT
ippsHashMethod_SHA384
ippsHashMethod_SHA384_NI
ippsHashMethod_SHA384_TT
ippsHashMethod_SHA512_256
ippsHashMethod_SHA512_256_NI
ippsHashMethod_SHA512_256_TT
ippsHashMethod_SHA512_224
ippsHashMethod_SHA512_224_NI
ippsHashMethod_SHA512_224_TT
ippsHashMethod_SHA3_224
ippsHashMethod_SHA3_256
ippsHashMethod_SHA3_384
ippsHashMethod_SHA3_512
ippsHashMethod_SHAKE128
ippsHashMethod_SHAKE256
ippsHashMethodGetInfo
ippsHashGetSize_rmf
ippsHashGetSizeOptimal_rmf
ippsHashInit_rmf
ippsHashPack_rmf
ippsHashUnpack_rmf
ippsHashDuplicate_rmf
ippsHashUpdate_rmf
ippsHashGetTag_rmf
ippsHashFinal_rmf
ippsHashMessage_rmf
ippsHashGetInfo_rmf
ippsHashSqueeze_rmf
ippsMGF
ippsMGF1_rmf
ippsMGF2_rmf
ippsHMAC_GetSize
ippsHMAC_Init
ippsHMAC_Pack
ippsHMAC_Unpack
ippsHMAC_Duplicate
ippsHMAC_Update
ippsHMAC_Final
ippsHMAC_GetTag
ippsHMAC_Message
ippsHMACGetSize_rmf
ippsHMACInit_rmf
ippsHMACPack_rmf
ippsHMACUnpack_rmf
ippsHMACDuplicate_rmf
ippsHMACUpdate_rmf
ippsHMACFinal_rmf
ippsHMACGetTag_rmf
ippsHMACMessage_rmf
ippsBigNumGetSize
ippsBigNumInit
ippsCmpZero_BN
ippsCmp_BN
ippsGetSize_BN
ippsSet_BN
ippsGet_BN
ippsRef_BN
ippsExtGet_BN
ippsAdd_BN
ippsSub_BN
ippsMul_BN
ippsMAC_BN_I
ippsDiv_BN
ippsMod_BN
ippsGcd_BN
ippsModInv_BN
ippsSetOctString_BN
ippsGetOctString_BN
ippsMontGetSize
ippsMontInit
ippsMontSet
ippsMontGet
ippsMontForm
ippsMontMul
ippsMontExp
ippsHashDRBG_EntropyInputCtxGetSize
ippsHashDRBG_EntropyInputCtxInit
ippsHashDRBG_GetSize
ippsHashDRBG_Init
ippsHashDRBG_Instantiate
ippsHashDRBG_Uninstantiate
ippsHashDRBG_Reseed
ippsHashDRBG_Gen
ippsHashDRBG_GenBN
ippsHashDRBG_InstantiateTest
ippsHashDRBG_ReseedTest
ippsHashDRBG_GenTest
ippsPRNGGetSize
ippsPRNGInit
ippsPRNGSetModulus
ippsPRNGSetH0
ippsPRNGSetAugment
ippsPRNGSetSeed
ippsPRNGGetSeed
ippsPRNGen
ippsPRNGen_BN
ippsPRNGenRDRAND
ippsPRNGenRDRAND_BN
ippsTRNGenRDSEED
ippsTRNGenRDSEED_BN
ippsPrimeGetSize
ippsPrimeInit
ippsPrimeGen
ippsPrimeTest
ippsPrimeGen_BN
ippsPrimeTest_BN
ippsPrimeGet
ippsPrimeGet_BN
ippsPrimeSet
ippsPrimeSet_BN
ippsRSA_GetSizePublicKey
ippsRSA_InitPublicKey
ippsRSA_SetPublicKey
ippsRSA_SetPublicKeyPart
ippsRSA_GetPublicKey
ippsRSA_GetSizePrivateKeyType1
ippsRSA_InitPrivateKeyType1
ippsRSA_SetPrivateKeyType1
ippsRSA_SetPrivateKeyType1Part
ippsRSA_GetPrivateKeyType1
ippsRSA_GetSizePrivateKeyType2
ippsRSA_InitPrivateKeyType2
ippsRSA_SetPrivateKeyType2
ippsRSA_SetPrivateKeyType2Part
ippsRSA_GetPrivateKeyType2
ippsRSA_GetBufferSizePublicKey
ippsRSA_GetBufferSizePrivateKey
ippsRSA_Encrypt
ippsRSA_Decrypt
ippsRSA_GenerateKeys
ippsRSA_ValidateKeys
ippsRSAEncrypt_OAEP
ippsRSADecrypt_OAEP
ippsRSAEncrypt_OAEP_rmf
ippsRSADecrypt_OAEP_rmf
ippsRSAEncrypt_PKCSv15
ippsRSADecrypt_PKCSv15
ippsRSASign_PSS
ippsRSAVerify_PSS
ippsRSASign_PSS_rmf
ippsRSAVerify_PSS_rmf
ippsRSASign_PKCS1v15
ippsRSAVerify_PKCS1v15
ippsRSASign_PKCS1v15_rmf
ippsRSAVerify_PKCS1v15_rmf
ippsDLGetResultString
ippsDLPGetSize
ippsDLPInit
ippsDLPPack
ippsDLPUnpack
ippsDLPSet
ippsDLPGet
ippsDLPSetDP
ippsDLPGetDP
ippsDLPGenKeyPair
ippsDLPPublicKey
ippsDLPValidateKeyPair
ippsDLPSetKeyPair
ippsDLPSignDSA
ippsDLPVerifyDSA
ippsDLPSharedSecretDH
ippsDLPGenerateDSA
ippsDLPValidateDSA
ippsDLPGenerateDH
ippsDLPValidateDH
ippsECCGetResultString
ippsGFpGetSize
ippsGFpInitArbitrary
ippsGFpInitFixed
ippsGFpInit
ippsGFpMethod_p192r1
ippsGFpMethod_p224r1
ippsGFpMethod_p256r1
ippsGFpMethod_p384r1
ippsGFpMethod_p521r1
ippsGFpMethod_p256sm2
ippsGFpMethod_p256bn
ippsGFpMethod_p256
ippsGFpMethod_pArb
ippsGFpxGetSize
ippsGFpxInit
ippsGFpxInitBinomial
ippsGFpxMethod_binom2_epid2
ippsGFpxMethod_binom3_epid2
ippsGFpxMethod_binom2
ippsGFpxMethod_binom3
ippsGFpxMethod_binom
ippsGFpxMethod_com
ippsGFpScratchBufferSize
ippsGFpElementGetSize
ippsGFpElementInit
ippsGFpSetElement
ippsGFpSetElementRegular
ippsGFpSetElementOctString
ippsGFpSetElementRandom
ippsGFpSetElementHash
ippsGFpSetElementHash_rmf
ippsGFpCpyElement
ippsGFpGetElement
ippsGFpGetElementOctString
ippsGFpCmpElement
ippsGFpIsZeroElement
ippsGFpIsUnityElement
ippsGFpConj
ippsGFpNeg
ippsGFpInv
ippsGFpSqrt
ippsGFpSqr
ippsGFpAdd
ippsGFpSub
ippsGFpMul
ippsGFpExp
ippsGFpMultiExp
ippsGFpAdd_PE
ippsGFpSub_PE
ippsGFpMul_PE
ippsGFpGetInfo
ippsGFpECGetSize
ippsGFpECInit
ippsGFpECSet
ippsGFpECSetSubgroup
ippsGFpECInitStd128r1
ippsGFpECInitStd128r2
ippsGFpECInitStd192r1
ippsGFpECInitStd224r1
ippsGFpECInitStd256r1
ippsGFpECInitStd384r1
ippsGFpECInitStd521r1
ippsGFpECInitStdSM2
ippsGFpECInitStdBN256
ippsGFpECBindGxyTblStd192r1
ippsGFpECBindGxyTblStd224r1
ippsGFpECBindGxyTblStd256r1
ippsGFpECBindGxyTblStd384r1
ippsGFpECBindGxyTblStd521r1
ippsGFpECBindGxyTblStdSM2
ippsGFpECGet
ippsGFpECGetSubgroup
ippsGFpECGetInfo_GF
ippsGFpECScratchBufferSize
ippsGFpECVerify
ippsGFpECPointGetSize
ippsGFpECPointInit
ippsGFpECSetPointAtInfinity
ippsGFpECSetPoint
ippsGFpECSetPointRegular
ippsGFpECSetPointRandom
ippsGFpECMakePoint
ippsGFpECSetPointHash
ippsGFpECSetPointHash_rmf
ippsGFpECSetPointHashBackCompatible
ippsGFpECSetPointHashBackCompatible_rmf
ippsGFpECGetPoint
ippsGFpECGetPointRegular
ippsGFpECSetPointOctString
ippsGFpECGetPointOctString
ippsGFpECTstPoint
ippsGFpECTstPointInSubgroup
ippsGFpECCpyPoint
ippsGFpECCmpPoint
ippsGFpECNegPoint
ippsGFpECAddPoint
ippsGFpECMulPoint
ippsGFpECPrivateKey
ippsGFpECPublicKey
ippsGFpECTstKeyPair
ippsGFpECSharedSecretDH
ippsGFpECSharedSecretDHC
ippsGFpECSignDSA
ippsGFpECVerifyDSA
ippsGFpECUserIDHashSM2
ippsGFpECMessageRepresentationSM2
ippsGFpECSignNR
ippsGFpECVerifyNR
ippsGFpECSignSM2
ippsGFpECVerifySM2
ippsGFpECKeyExchangeSM2_GetSize
ippsGFpECKeyExchangeSM2_Init
ippsGFpECKeyExchangeSM2_Setup
ippsGFpECKeyExchangeSM2_SharedKey
ippsGFpECKeyExchangeSM2_Confirm
ippsGFpECEncryptSM2_Ext_EncMsgSize
ippsGFpECEncryptSM2_Ext
ippsGFpECDecryptSM2_Ext_DecMsgSize
ippsGFpECDecryptSM2_Ext
ippsGFpECESGetSize_SM2
ippsGFpECESInit_SM2
ippsGFpECESSetKey_SM2
ippsGFpECESStart_SM2
ippsGFpECESEncrypt_SM2
ippsGFpECESDecrypt_SM2
ippsGFpECESFinal_SM2
ippsGFpECESGetBuffersSize_SM2
ippsXMSSVerify
ippsXMSSSetPublicKeyState
ippsXMSSSetSignatureState
ippsXMSSSignatureStateGetSize
ippsXMSSPublicKeyStateGetSize
ippsXMSSBufferGetSize
ippsXMSSVerifyBufferGetSize
ippsLMSBufferGetSize
ippsLMSVerifyBufferGetSize
ippsLMSSignatureStateGetSize
ippsLMSPublicKeyStateGetSize
ippsLMSSetPublicKeyState
ippsLMSSetSignatureState
ippsLMSVerify
ippsLMSKeyGenBufferGetSize
ippsLMSSignBufferGetSize
ippsLMSPrivateKeyStateGetSize
ippsLMSInitSignature
ippsLMSInitKeyPair
ippsLMSKeyGen
ippsLMSSign
ippsHKDF
ippsHKDF_extract
ippsHKDF_expand
ippsXMSSKeyGenBufferGetSize
ippsXMSSSignBufferGetSize
ippsXMSSPrivateKeyStateGetSize
ippsXMSSInitKeyPair
ippsXMSSInitSignature
ippsXMSSKeyGen
ippsXMSSSign
ippsMLKEM_GetSize
ippsMLKEM_KeyGenBufferGetSize
ippsMLKEM_EncapsBufferGetSize
ippsMLKEM_DecapsBufferGetSize
ippsMLKEM_GetInfo
ippsMLKEM_Init
ippsMLKEM_KeyGen
ippsMLKEM_Encaps
ippsMLKEM_Decaps
ippsPBKDF2_PKCS5v2
ippsMLDSA_GetSize
ippsMLDSA_GetInfo
ippsMLDSA_Init
ippsMLDSA_KeyGenBufferGetSize
ippsMLDSA_SignBufferGetSize
ippsMLDSA_VerifyBufferGetSize
ippsMLDSA_KeyGen
ippsMLDSA_Sign
ippsMLDSA_Verify

ippcp_is_fips_approved_func

fips_selftest_ippsAESEncryptDecrypt_get_size

fips_selftest_ippsAESEncryptCBC
fips_selftest_ippsAESDecryptCBC

fips_selftest_ippsAESEncryptCBC_CS1
fips_selftest_ippsAESEncryptCBC_CS2
fips_selftest_ippsAESEncryptCBC_CS3

fips_selftest_ippsAESDecryptCBC_CS1
fips_selftest_ippsAESDecryptCBC_CS2
fips_selftest_ippsAESDecryptCBC_CS3

fips_selftest_ippsAESEncryptCFB
fips_selftest_ippsAESDecryptCFB

fips_selftest_ippsAESEncryptOFB
fips_selftest_ippsAESDecryptOFB

fips_selftest_ippsAESEncryptCTR
fips_selftest_ippsAESDecryptCTR

fips_selftest_ippsAESEncryptDecryptCCM_get_size
fips_selftest_ippsAES_CCMEncrypt
fips_selftest_ippsAES_CCMDecrypt

fips_selftest_ippsAES_GCM_get_size
fips_selftest_ippsAES_GCMEncrypt
fips_selftest_ippsAES_GCMDecrypt

fips_selftest_ippsAES_CMAC_get_size
fips_selftest_ippsAES_CMACUpdate

fips_selftest_ippsRSAEncryptDecrypt_OAEP_rmf_get_size_keys
fips_selftest_ippsRSAEncryptDecrypt_OAEP_rmf_get_size
fips_selftest_ippsRSAEncrypt_OAEP_rmf
fips_selftest_ippsRSADecrypt_OAEP_rmf

fips_selftest_ippsHash_rmf_get_size
fips_selftest_ippsHashUpdate_rmf
fips_selftest_ippsHashMessage_rmf

fips_selftest_ippsHMAC_rmf_get_size
fips_selftest_ippsHMACUpdate_rmf
fips_selftest_ippsHMACMessage_rmf

fips_selftest_ippsRSASign_PKCS1v15_rmf_get_size_keys
fips_selftest_ippsRSASign_PKCS1v15_rmf_get_size
fips_selftest_ippsRSAVerify_PKCS1v15_rmf_get_size_keys
fips_selftest_ippsRSAVerify_PKCS1v15_rmf_get_size
fips_selftest_ippsRSASign_PKCS1v15_rmf
fips_selftest_ippsRSAVerify_PKCS1v15_rmf

fips_selftest_ippsRSASignVerify_PSS_rmf_get_size_keys
fips_selftest_ippsRSASignVerify_PSS_rmf_get_size
fips_selftest_ippsRSASign_PSS_rmf
fips_selftest_ippsRSAVerify_PSS_rmf
fips_selftest_ippsRSA_GenerateKeys

fips_selftest_ippsGFpECSignVerifyDSA_get_size_GFp_buff
fips_selftest_ippsGFpECSignVerifyDSA_get_size_GFpEC_buff
fips_selftest_ippsGFpECSignVerifyDSA_get_size_data_buff
fips_selftest_ippsGFpECSignDSA
fips_selftest_ippsGFpECVerifyDSA
fips_selftest_ippsGFpECPublicKey
fips_selftest_ippsGFpECPrivateKey
fips_selftest_ippsGFpECSharedSecretDH
fips_selftest_ippsLMSVerify_get_size
fips_selftest_ippsLMSVerify

fips_selftest_ippsMLDSA_Sign_get_size_data_buff
fips_selftest_ippsMLDSA_Verify_get_size_data_buff
fips_selftest_ippsMLDSA_KeyGen_get_size_data_buff
fips_selftest_ippsMLDSA_Sign_get_size
fips_selftest_ippsMLDSA_Verify_get_size
fips_selftest_ippsMLDSA_KeyGen_get_size
fips_selftest_ippsMLDSA_Sign
fips_selftest_ippsMLDSA_Verify
fips_selftest_ippsMLDSA_KeyGen
